unit UInstructionWriter;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, UCpu;

type

  { TInstructionWriter }

  TInstructionWriter = class
  private
    DataSizeLast: TBitWidth;
    DataSizePrefix: TBitWidth;
    AddrSizeLast: TBitWidth;
    AddrSizePrefix: TBitWidth;
    procedure PrefixBegin;
    procedure PrefixEnd;
  public
    Cpu: TCpu;
    IP: Integer;
    DataSize: TBitWidth;
    AddrSize: TBitWidth;
    procedure Init;
    procedure Write8(Value: Byte);
    procedure Write16(Value: Word);
    procedure Write32(Value: DWord);
    procedure Write64(Value: QWord);
    procedure WriteAddress(Value: TAddress);
    procedure WriteAddressSigned(Value: TAddressSigned);
    procedure WriteData(Value: QWord);
    procedure WriteString(Text: string);
    procedure Nop;
    procedure Halt;
    procedure Loadi(Reg: TRegIndex; Value: QWord);
    procedure Increment(Reg: TRegIndex);
    procedure Decrement(Reg: TRegIndex);
    procedure Jump(Addr: QWord);
    procedure JumpNotZero(Addr: QWord);
    procedure JumpZero(Addr: QWord);
    procedure JumpRel(Addr: QWord);
    procedure JumpRelNotZero(Addr: QWord);
    procedure JumpRelZero(Addr: QWord);
    procedure Push(Reg: TRegIndex);
    procedure Pop(Reg: TRegIndex);
    procedure DataPrefix8;
    procedure DataPrefix16;
    procedure DataPrefix32;
    procedure DataPrefix64;
    procedure AddrPrefix8;
    procedure AddrPrefix16;
    procedure AddrPrefix32;
    procedure AddrPrefix64;
    procedure StoreMem(RegAddr, RegSrc: TRegIndex);
    procedure LoadMem(RegDst, RegAddr: TRegIndex);
    procedure Output(Port: TAddress; Reg: TRegIndex);
    procedure Input(Reg: TRegIndex; Port: TAddress);
    procedure Test(Reg: TRegIndex);
  end;

  { TAssembler }

  TAssembler = class
    Source: TStringList;
    Dest: array of Byte;
    procedure Compile;
  end;

  TDisassembler = class

  end;

const
  OpcodeName: array[TOpcode] of string = ('NOP', 'HALT', 'LD', 'LDI', 'JP', 'JPZ',
    'JPNZ', 'JR', 'JRZ', 'JRNZ', 'NEG', 'CLR', 'LDM', 'STM', 'EX', 'PUSH', 'POP',
    'CALL', 'RET', 'ADD', 'ADDI', 'SUB', 'SUBI', 'INC', 'DEC', 'IN', 'OUT',
    'SHL', 'SHR', 'DP8', 'DP16', 'DP32', 'DP64', 'DS', 'AS', 'TEST', 'AND', 'OR',
    'XOR', 'LDDR', 'LDIR', 'MUL', 'DIV', 'MOD', 'AP8', 'AP16', 'AP32', 'AP64');

implementation

{ TAssembler }

procedure TAssembler.Compile;
begin

end;

{ TInstructionWriter }

procedure TInstructionWriter.Nop;
begin
  PrefixBegin;
  Write8(Byte(opNop));
  PrefixEnd;
end;

procedure TInstructionWriter.Halt;
begin
  PrefixBegin;
  Write8(Byte(opHalt));
  PrefixEnd;
end;

procedure TInstructionWriter.Loadi(Reg: TRegIndex; Value: QWord);
begin
  PrefixBegin;
  Write8(Byte(opLoadi));
  Write8(Reg);
  WriteData(Value);
  PrefixEnd;
end;

procedure TInstructionWriter.Increment(Reg: TRegIndex);
begin
  PrefixBegin;
  Write8(Byte(opInc));
  Write8(Reg);
  PrefixEnd;
end;

procedure TInstructionWriter.Decrement(Reg: TRegIndex);
begin
  PrefixBegin;
  Write8(Byte(opDec));
  Write8(Reg);
  PrefixEnd;
end;

procedure TInstructionWriter.Jump(Addr: QWord);
begin
  PrefixBegin;
  Write8(Byte(opJump));
  WriteAddress(Addr);
  PrefixEnd;
end;

procedure TInstructionWriter.JumpNotZero(Addr: QWord);
begin
  PrefixBegin;
  Write8(Byte(opJumpNotZero));
  WriteAddress(Addr);
  PrefixEnd;
end;

procedure TInstructionWriter.JumpZero(Addr: QWord);
begin
  PrefixBegin;
  Write8(Byte(opJumpZero));
  WriteAddress(Addr);
  PrefixEnd;
end;

procedure TInstructionWriter.JumpRel(Addr: QWord);
var
  NextIP: QWord;
begin
  PrefixBegin;
  NextIP := IP + 1 + BitWidthBytes[Cpu.AddrSize];
  Write8(Byte(opJumpRel));
  WriteAddressSigned(Int64(Addr) - Int64(NextIP));
  PrefixEnd;
end;

procedure TInstructionWriter.JumpRelNotZero(Addr: QWord);
var
  NextIP: QWord;
begin
  PrefixBegin;
  NextIP := IP + 1 + BitWidthBytes[Cpu.AddrSize];
  Write8(Byte(opJumpRelNotZero));
  WriteAddressSigned(Int64(Addr) - Int64(NextIP));
  PrefixEnd;
end;

procedure TInstructionWriter.JumpRelZero(Addr: QWord);
var
  NextIP: QWord;
begin
  PrefixBegin;
  NextIP := IP + 1 + BitWidthBytes[Cpu.AddrSize];
  Write8(Byte(opJumpRelZero));
  WriteAddressSigned(Int64(Addr) - Int64(NextIP));
  if AddrSizeLast <> bwNone then AddrSize := AddrSizeLast;
  PrefixEnd;
end;

procedure TInstructionWriter.Push(Reg: TRegIndex);
begin
  PrefixBegin;
  Write8(Byte(opPush));
  Write8(Reg);
  PrefixEnd;
end;

procedure TInstructionWriter.Pop(Reg: TRegIndex);
begin
  PrefixBegin;
  Write8(Byte(opPop));
  Write8(Reg);
  PrefixEnd;
end;

procedure TInstructionWriter.DataPrefix8;
begin
  DataSizePrefix := bw8;
  Write8(Byte(opDataPrefix8));
end;

procedure TInstructionWriter.DataPrefix16;
begin
  DataSizePrefix := bw16;
  Write8(Byte(opDataPrefix16));
end;

procedure TInstructionWriter.DataPrefix32;
begin
  DataSizePrefix := bw32;
  Write8(Byte(opDataPrefix32));
end;

procedure TInstructionWriter.DataPrefix64;
begin
  DataSizePrefix := bw64;
  Write8(Byte(opDataPrefix64));
end;

procedure TInstructionWriter.AddrPrefix8;
begin
  AddrSizePrefix := bw8;
  Write8(Byte(opAddrPrefix8));
end;

procedure TInstructionWriter.AddrPrefix16;
begin
  AddrSizePrefix := bw16;
  Write8(Byte(opAddrPrefix16));
end;

procedure TInstructionWriter.AddrPrefix32;
begin
  AddrSizePrefix := bw32;
  Write8(Byte(opAddrPrefix32));
end;

procedure TInstructionWriter.AddrPrefix64;
begin
  AddrSizePrefix := bw64;
  Write8(Byte(opAddrPrefix64));
end;

procedure TInstructionWriter.StoreMem(RegAddr, RegSrc: TRegIndex);
begin
  Write8(Byte(opStoreMem));
  Write8(RegAddr);
  Write8(RegSrc);
end;

procedure TInstructionWriter.LoadMem(RegDst, RegAddr: TRegIndex);
begin
  PrefixBegin;
  Write8(Byte(opLoadMem));
  Write8(RegDst);
  Write8(RegAddr);
  PrefixEnd;
end;

procedure TInstructionWriter.Output(Port: TAddress; Reg: TRegIndex);
begin
  PrefixBegin;
  Write8(Byte(opOut));
  WriteAddress(Port);
  Write8(Reg);
  PrefixEnd;
end;

procedure TInstructionWriter.Input(Reg: TRegIndex; Port: TAddress);
begin
  PrefixBegin;
  Write8(Byte(opIn));
  Write8(Reg);
  WriteAddress(Port);
  PrefixEnd;
end;

procedure TInstructionWriter.Test(Reg: TRegIndex);
begin
  PrefixBegin;
  Write8(Byte(opTest));
  Write8(Reg);
  Prefixend;
end;

procedure TInstructionWriter.WriteString(Text: string);
var
  I: Integer;
begin
  for I := 1 to Length(Text) do
    Write8(Ord(Text[I]));
end;

procedure TInstructionWriter.PrefixBegin;
begin
  if DataSizePrefix <> bwNone then begin
    DataSizeLast := DataSize;
    DataSize := DataSizePrefix;
    DataSizePrefix := bwNone;
  end;
  if AddrSizePrefix <> bwNone then begin
    AddrSizeLast := AddrSize;
    AddrSize := AddrSizePrefix;
    AddrSizePrefix := bwNone;
  end;
end;

procedure TInstructionWriter.PrefixEnd;
begin
  if DataSizeLast <> bwNone then begin
    DataSize := DataSizeLast;
    DataSizeLast := bwNone;
  end;
  if AddrSizeLast <> bwNone then begin
    AddrSize := AddrSizeLast;
    AddrSizeLast := bwNone;
  end;
end;

procedure TInstructionWriter.Init;
begin
  DataSize := Cpu.DataSize;
  AddrSize := Cpu.AddrSize;
end;

procedure TInstructionWriter.Write8(Value: Byte);
begin
  PByte(Cpu.Memory + IP)^ := Value;
  Inc(IP, SizeOf(Byte));
end;

procedure TInstructionWriter.Write16(Value: Word);
begin
  PWord(Cpu.Memory + IP)^ := Value;
  Inc(IP, SizeOf(Word));
end;

procedure TInstructionWriter.Write32(Value: DWord);
begin
  PDWord(Cpu.Memory + IP)^ := Value;
  Inc(IP, SizeOf(DWord));
end;

procedure TInstructionWriter.Write64(Value: QWord);
begin
  PQWord(Cpu.Memory + IP)^ := Value;
  Inc(IP, SizeOf(QWord));
end;

procedure TInstructionWriter.WriteAddress(Value: TAddress);
begin
  case AddrSize of
    bw8: Write8(Value);
    bw16: Write16(Value);
    bw32: Write32(Value);
    bw64: Write64(Value);
  end;
end;

procedure TInstructionWriter.WriteAddressSigned(Value: TAddressSigned);
begin
  case AddrSize of
    bw8: Write8(Byte(Value));
    bw16: Write16(Word(Value));
    bw32: Write32(DWord(Value));
    bw64: Write64(QWord(Value));
  end;
end;

procedure TInstructionWriter.WriteData(Value: QWord);
begin
  case DataSize of
    bw8: Write8(Value);
    bw16: Write16(Value);
    bw32: Write32(Value);
    bw64: Write64(Value);
  end;
end;

end.

